import network
import socket
import time
from machine import Pin
import uasyncio as asyncio
import re
import gc
import epd_small as epd

gc.collect()

ledocka = Pin("LED", Pin.OUT) #Pico W
btn0 = Pin(16, Pin.IN, Pin.PULL_DOWN)
btn1 = Pin(17, Pin.IN, Pin.PULL_DOWN)
btn2 = Pin(18, Pin.IN, Pin.PULL_DOWN)
tlacitka=[False,False,False]
aktualizujem=False

displej_w=296
displej_h=128

ssid = 'Nazov Wifi siete'
passwd = 'heslo'

btn_files=['sprava0.txt','sprava1.txt','sprava2.txt']
html = ""

wlan = network.WLAN(network.STA_IF)

def vypis_ifconfig(ifconf):
    print("ipcka:",ifconf[0])
    print("subnet maska:",ifconf[1])
    print("gateway:",ifconf[2])
    print("DNS:",ifconf[3])

# Desifruje url enkodovanie
# Upravene z: https://github.com/gurgleapps/pico-web-server-control
def url_decode(encoded_str):
    decoded_str = ""
    i = 0
    while i < len(encoded_str):
        if encoded_str[i] == '%':
            hex_code = encoded_str[i + 1:i + 3]
            char = chr(int(hex_code, 16))
            decoded_str += char
            i += 3
        else:
            decoded_str += encoded_str[i]
            i += 1
    decoded_str = decoded_str.replace('+',' ')
    return decoded_str

def parse_form_data(post_data):
    pary=post_data.split('&')
    n=len(pary)
    print("n:",n)
    form_data=[""]*n
    for i in range(n):
        form_data[i] = url_decode(pary[i].split('=')[1]) 
    return form_data

def zapis_subor(nazov,sprava):
    f=open(nazov,'w')
    f.write(sprava)
    f.flush()
    f.close()

def citaj_subor(nazov):
    obsah=""
    f=open(nazov,'r',encoding="utf-8")
    obsah=f.read()
    f.close()
    return obsah

def aktualizuj_displej(sprava):
    global aktualizujem
    if aktualizujem:
        return
    aktualizujem=True
    eped = epd.EPD_2in9_Landscape()

    dl=len(sprava)*8
    x=(displej_w-dl)//2
    y=(displej_h-8)//2
    off=8

    eped.Clear(0xff)
    eped.fill(0xff)
    eped.text(sprava, x, y, 0x00)
    eped.rect(x-off,y-off,dl+(off*2),8+(off*2),0x00)
    eped.display(eped.buffer)
    eped.delay_ms(2000)

    eped.init()
    # eped.Clear(0xff)
    eped.delay_ms(2000)
    print("sleep")
    eped.sleep()
    aktualizujem=False

def pripoj_sa():
    if wlan.isconnected():
        print("Uz som pripojeny??")
    wlan.active(True)
    wlan.config(pm=0xa11140)
    wlan.connect(ssid, passwd)
    
    max_cakaj=15
    while max_cakaj > 0:
        if wlan.status() < 0 or wlan.status() >=3:
            break
        max_cakaj-=1
        print("Pripajam sa na siet...")
        time.sleep(1)
    
    if wlan.status()!=3:
        raise RuntimeError('Pripojenie zlyhalo!')
    else:
        print("Pripojene")
        status = wlan.ifconfig()
        vypis_ifconfig(status)
        
async def obsluz_klienta(reader, writer):
    gc.collect()
    print("Klient sa pripojil")
    headre=[]
    metoda=""
    odp=""
    content_length=0
    post_data_raw=None
    post_data=None
    form_data=None
    
    while True:
        lajna = await reader.readline()
        lajna = lajna.decode("utf-8").strip()
        if lajna == "":
            break
        headre.append(lajna)
    request_raw = str("\r\n".join(headre))
    print(request_raw)
    
    request_pattern = re.compile(r"(GET|POST)\s+([^\s]+)\s+HTTP")
    match = request_pattern.search(request_raw)
    
    metoda = match.group(1)
    
    if metoda == "POST":
        content_length_pattern = re.compile(r"Content-Length:\s+(\d+)")
        match = content_length_pattern.search(request_raw)
        if match:
            content_length = int(match.group(1))
            print("content-length: "+str(content_length))
        if content_length>0:
            odp="Správa bola úspešne zapísaná!"
            post_data_raw = await reader.readexactly(content_length)
            # print("Raw POST data:",post_data_raw)
            
            post_data=post_data_raw.decode('utf-8')
            form_data=parse_form_data(post_data)
            
            print("Form data:",form_data)
            sprava=form_data[0].replace('\n', '')
            index=form_data[1]
            
            if index != "":
                index=int(index)
                zapis_subor(btn_files[index],sprava)
                aktualizuj_displej(sprava)
            else:
                aktualizuj_displej(sprava)
            
    odpoved = html % odp
    writer.write('HTTP/1.0 200 OK\r\nContent-type: text/html\r\n\r\n')
    writer.write(odpoved)

    await writer.drain()
    await writer.wait_closed()
    print("Klient sa odpojil")
    
    
def spravca0(pin):
    global tlacitka
    tlacitka[0]=True
        
def spravca1(pin):
    global tlacitka
    tlacitka[1]=True
    
def spravca2(pin):
    global tlacitka
    tlacitka[2]=True
        
def spracuj_tlacitka():
    global tlacitka
    for i in range(len(tlacitka)):
        if tlacitka[i]:
            sprava=citaj_subor(btn_files[i])
            aktualizuj_displej(sprava)
            tlacitka[i]=False


html=citaj_subor('web.html')

async def main():
    btn0.irq(trigger=Pin.IRQ_RISING, handler=spravca0)
    btn1.irq(trigger=Pin.IRQ_RISING, handler=spravca1)
    btn2.irq(trigger=Pin.IRQ_RISING, handler=spravca2)
    print("Pripajam sa na siet...")
    pripoj_sa()

    print("Startujem webserver...")
    server = asyncio.start_server(obsluz_klienta, "0.0.0.0", 80)
    asyncio.create_task(server)
    while True:
        spracuj_tlacitka()
        ledocka.on()
        await asyncio.sleep(0.5)
        ledocka.off()
        await asyncio.sleep(5)

try:
    asyncio.run(main())
except KeyboardInterrupt as e:
    loop = asyncio.get_event_loop()
    loop.stop()
    ledocka.off()
    wlan.disconnect()
    wlan.active(False)
    print(wlan.active())
finally:
    asyncio.new_event_loop()